;****************************************************************************
;*                                                                          *
;* cane - A softdog killer                                                  *
;* Copyright (C) 1996, by Li Wen_Xian                                       *
;* All rights reserved.                                                     *
;*                                                                          *
;* This file is: INT386.ASM, the VM interrupt handlers                      *
;*                                                                          *
;****************************************************************************

INCLUDE         EQUMAC.INC
INT386_FILE     EQU     1
INCLUDE         EXTERNAL.INC

;fake segment for far ret interrupts
;(this segment has no descriptor in GDT/LDT)
QISR            SEGMENT PARA PUBLIC 'CODE16' USE16
                ASSUME  CS:QISR
;push   sacrifical words for IRET to eat.
;PL0 stack controls return anyway
QIRET:
                PUSH    0
                PUSH    0
                PUSH    0
                IRET
QISR            ENDS

;IDT    segment
IDTABLE         SEGMENT
IDTBEG          EQU             $
                DQ              TOPINT+1 DUP (0)
IDTEND          EQU             $
IDTABLE         ENDS

;ISR segment
DEFINT          MACRO           N
INT&N           LABEL           FAR
                DB              6AH
                DB              N
                JMP             NEAR PTR INTDUMP
                ENDM

ISR             SEGMENT
                ASSUME  CS:ISR
ISRBEG          EQU             $

;This code defines interrupt handlers form 0 to TOPINT
;(TOPINT is defined in EQUMAC.INC)
;The interrupt code assumes these routines are the same
;size -- even if the assembler tells you that some of
;them can be compatcted with an override -- don't do it
;Also, don't let the assembler make multiple passes to
;compact this code automatically.
INTNO           =       0
                REPT    TOPINT+1
                DEFINT  %INTNO
INTNO           =       INTNO+1
                ENDM
;Debug dump messages
MESSAGEA        DB      'INT=',0
STKM            DB      'Stack dump:',0
TASKM           DB      'TR=',0
RTABLE          DB      'G'
                DB      'F'
                DB      'D'
                DB      'E'
GTABLE          DB      'DISIBPSPBXDXCXAX'
MEMMESS         DB      'Memory Dump:',0

;****************************************************************************
;All interrupts come here                                                   *
;We check for the interrupt # pushed on the stack and                       *
;vector accordingly. This adds some interrupt latency,                      *
;but simplifies IDT construction.                                           *
;****************************************************************************

INTDUMP         LABEL   NEAR

                PUSHAD
                PUSH    ES
                PUSH    FS
                PUSH    GS

                MOV     AX,SEL_DATA0     ;ES  --->  SEL_DATA0
                MOV     ES,AX
                MOV     AX,SEL_DATA      ;FS  --->  SEL_DATA
                MOV     FS,AX
                MOV     ECX,[ESP+2CH]
                AND     ECX,000000FFH
                MOV     [ESP+2CH],ECX    ;interrupt no.

                MOV     ECX,[ESP+2CH]    ;INT No
                CMP     ECX,00H          ;DIVIDE BY 0 INTERRUPT
                JZ      INT00H
                CMP     ECX,01H          ;STEP
                JZ      INTNN
                CMP     ECX,02H          ;NMI
                JZ      INTNN
                CMP     ECX,03H          ;BREAKPOINT INTERRUPT
                JZ      INTNN
                CMP     ECX,04H          ;OVERFLOW INTERRUPT
                JZ      INTNN
                CMP     ECX,08H
                JB      NOTINT08_0F
                CMP     ECX,0FH
                JA      NOTINT08_0F
                SUB     CL,08H
                MOV     AL,0BH    ;write OCW3
                OUT     20H,AL
                IN      AL,20H    ;READ ISR
                SHR     AL,CL
                AND     AL,01H
                JZ      NOTHARDINT
INTNN:
                MOV     EAX,[ESP+38H] ;OLD FLAGS
                AND     EAX,0FFFFFEFFH ;CLEAR TF
                MOV     [ESP+38H],EAX

                MOV     EAX,[ESP+40H] ;OLD SS
                SHL     EAX,4
                MOV     EBX,[ESP+3CH] ;OLD ESP
                SUB     EBX,6
                AND     EBX,0FFFFH
                ADD     EBX,EAX

                MOV     ECX,[ESP+38H] ;OLD FLAGS
                MOV     EAX,[ESP+34H] ;OLD CS
                AND     EAX,0FFFFH
                SHL     ECX,10H
                OR      EAX,ECX
                MOV     ES:[EBX+2],EAX
                MOV     EAX,[ESP+30H] ;OLD EIP
                MOV     ES:[EBX],AX

                MOV     EAX,[ESP+3CH] ;OLD ESP
                SUB     EAX,6
                AND     EAX,0FFFFH
                MOV     [ESP+3CH],EAX ;NEW ESP

                MOV     EAX,[ESP+38H] ;FLAGS
                AND     EAX,0FFFFFDFFH ;CLEAR IF
                MOV     [ESP+38H],EAX

                MOV     EBX,[ESP+2CH]  ;INT NO
                MOV     EAX,ES:[EBX*4]
                MOV     EDX,EAX
                SHR     EDX,10H
                AND     EAX,0FFFFH
                MOV     [ESP+30H],EAX
                MOV     [ESP+34H],EDX

                POP     GS
                POP     FS
                POP     ES
                POPAD
                ADD     ESP,4
                IRETD
INT00H:
                MOV     EAX,[ESP+38H] ;OLD FLAGS
                AND     EAX,0FFFFF606H ;CLEAR EXCEPT DF IF PF
                MOV     [ESP+38H],EAX
                JMP     INTNN

NOTINT08_0F:
                CMP     ECX,70H    ;8259A 2
                JB      NOTHARDINT
                CMP     ECX,77H
                JA      NOTHARDINT
                SUB     CL,70H
                MOV     AL,0BH   ;WRITE OCW3
                OUT     20H,AL
                IN      AL,20H   ;READ ISR
                SHR     AL,2
                AND     AL,01H
                JZ      NOTHARDINT
                MOV     AL,0BH   ;WRITE OCW3
                OUT     0A0H,AL
                IN      AL,0A0H   ;READ ISR
                SHR     AL,CL
                AND     AL,01H
                JZ      NOTHARDINT
                JMP     INTNN
NOTHARDINT:
                MOV     EAX,[ESP+34H] ;OLD CS
                SHL     EAX,4
                MOV     EBX,[ESP+30H] ;OLD EIP
                ADD     EBX,EAX
                MOV     EAX,ES:[EBX-4] ;the interrupt instruction
                AND     EAX,00FF0000H
                CMP     EAX,00CD0000H
                JNZ     NOTINTNN
                MOV     EAX,ES:[EBX-4]
                SHR     EAX,24
                CMP     EAX,[ESP+2CH]
                JNZ     NOTINTNN
                CMP     EAX,15H
                JZ      INT15H
                CMP     EAX,21H
                JZ      INT21H
                JMP     INTNN
INT21H:
                MOV     AX,WORD PTR [ESP+28H]
                CMP     AX,0FEFEH
                JZ      GETPSP
                CMP     AX,0FEFFH
                JZ      BACKTOREAL
                CMP     AX,0FEFDH
                JZ      GETBUFADDR
                CMP     AX,0FEFCH
                JZ      SETBUFADDR
                JMP     INTNN
SETBUFADDR:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EAX,[ESP+1CH]     ;OLD EBX
                MOV     GS:PORTDATAADDR,EAX
                ADD     EAX,2
                MOV     GS:CURBLOCKADDR,EAX
                MOV     WORD PTR [ESP+28H],0  ;set eax
                POP     GS
                POP     FS
                POP     ES
                POPAD
                ADD     ESP,4
                IRETD
GETBUFADDR:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EAX,GS:PORTDATAADDR
                MOV     [ESP+24H],EAX                ;ecx the address
                MOV     AX,0H
                MOV     WORD PTR [ESP+28H],AX
                POP     GS
                POP     FS
                POP     ES
                POPAD
                ADD     ESP,4
                IRETD
GETPSP:
                MOV     AX,FS:_PSP
                MOV     WORD PTR [ESP+28H],AX
                POP     GS
                POP     FS
                POP     ES
                POPAD
                ADD     ESP,4
                IRETD

BACKTOREAL:
                MOV     AL,20H
                CMP     FS:_PC386,0
                JNZ     SHORT NOTANAT1
                OUT     0A0H,AL
NOTANAT1:
                OUT     20H,AL          ;just in case hardware did it
                MOV     EAX,SEG _RTCSIP
                SHL     EAX,4
                MOV     EBX,OFFSET _RTCSIP
                ADD     EBX,EAX
                MOV     AX,WORD PTR [ESP+30H]    ;OLD IP
                MOV     WORD PTR ES:[EBX],AX
                MOV     AX,WORD PTR [ESP+34H]    ;OLD CS
                MOV     WORD PTR ES:[EBX+2],AX
                MOV     AX,WORD PTR [ESP+38H]    ;OLD FLAG
                MOV     WORD PTR ES:[EBX+4],AX
                MOV     AX,WORD PTR [ESP+44H]    ;OLD ES
                MOV     WORD PTR ES:[EBX+6],AX
                MOV     AX,WORD PTR [ESP+48H]    ;OLD DS
                MOV     WORD PTR ES:[EBX+8],AX
                MOV     AX,WORD PTR [ESP+40H]    ;OLD SS
                MOV     WORD PTR ES:[EBX+0AH],AX
                MOV     AX,WORD PTR [ESP+14H]    ;OLD BP
                MOV     WORD PTR ES:[EBX+0CH],AX
                MOV     AX,WORD PTR [ESP+3CH]    ;OLD SP
                MOV     WORD PTR ES:[EBX+0EH],AX
                MOV     AX,WORD PTR [ESP+0CH]    ;OLD DI
                MOV     WORD PTR ES:[EBX+10H],AX
                MOV     AX,WORD PTR [ESP+10H]    ;OLD SI
                MOV     WORD PTR ES:[EBX+12H],AX
                MOV     AX,WORD PTR [ESP+20H]    ;OLD DX
                MOV     WORD PTR ES:[EBX+14H],AX
                MOV     AX,WORD PTR [ESP+24H]    ;OLD CX
                MOV     WORD PTR ES:[EBX+16H],AX
                MOV     AX,WORD PTR [ESP+1CH]    ;OLD BX
                MOV     WORD PTR ES:[EBX+18H],AX
                MOV     AX,WORD PTR [ESP+28H]    ;OLD AX
                MOV     WORD PTR ES:[EBX+1AH],AX

                POP     GS
                POP     FS
                POP     ES
                POPAD
                ADD     ESP,4
                BACK2REAL
INT15H:
                MOV     EAX,[ESP+28H]    ;OLD EAX
                SHR     EAX,8
                AND     EAX,0FFH         ;OLD AH
                CMP     EAX,87H
                JZ      TRANSDATA
                CMP     EAX,88H
                JZ      MEMORYCHECK
                JMP     INTNN
TRANSDATA:
                MOV     AX,SEL_GDT       ;GS  --->  SEL_GDT
                MOV     GS,AX
                MOV     EAX,[ESP+44H]    ;OLD ES
                AND     EAX,0FFFFH
                SHL     EAX,4
                MOV     EBX,[ESP+10H]    ;OLD SI
                AND     EBX,0FFFFH
                ADD     EBX,EAX
                MOV     EAX,ES:[EBX+10H]   ;SOURCE
                MOV     GS:[SEL_15S],EAX
                MOV     EAX,ES:[EBX+14H]
                MOV     GS:[SEL_15S+4],EAX
                MOV     EAX,ES:[EBX+18H]   ;DESTINATION
                MOV     GS:[SEL_15D],EAX
                MOV     EAX,ES:[EBX+1CH]
                MOV     GS:[SEL_15D+4],EAX
                MOV     CX,WORD PTR [ESP+24H]    ;TRANS LEN
                MOVZX   ECX,CX
                PUSH    DS
                MOV     AX,SEL_15S
                MOV     DS,AX            ;SOURCE      TRANS
                MOV     AX,SEL_15D
                MOV     ES,AX            ;DESTINATION TRANS
                XOR     ESI,ESI
                XOR     EDI,EDI
                CLD
                REP     MOVSW
                POP     DS
                MOV     [ESP+28H],EAX
                POP     GS
                POP     FS
                POP     ES
                POPAD
                ADD     ESP,4
                IRETD
MEMORYCHECK:
                MOV     AL,31H
                OUT     70H,AL
                IN      AL,71H
                SHL     EAX,8
                MOV     AL,30H
                OUT     70H,AL
                IN      AL,71H
                AND     EAX,0FFFFH
                MOV     [ESP+28H],EAX
                POP     GS
                POP     FS
                POP     ES
                POPAD
                ADD     ESP,4
                IRETD
NOTINTNN:                                ; Is trap
                MOV     ECX,[ESP+2CH]
                CMP     ECX,0DH
                JNZ     NOTGP

                ;Is 13, general protect error
                MOV     EAX,[ESP+34H+4]  ;OLD CS
                AND     EAX,0FFFFH
                SHL     EAX,4
                MOV     EBX,[ESP+30H+4]  ;OLD EIP
                ADD     EBX,EAX
                MOV     EAX,ES:[EBX]     ;current instruction
                AND     EAX,0FFH
                CMP     EAX,0FH
                JNZ     NOTMOV           ;NOT mov instruction
                MOV     EAX,ES:[EBX]     ;current instruction
                SHR     EAX,8
                AND     EAX,0FFH
                CMP     EAX,22H
                JZ      SKIPINSTRUCTION
                CMP     EAX,23H
                JZ      SKIPINSTRUCTION
                CMP     EAX,26H
                JZ      SKIPINSTRUCTION
                CMP     EAX,20H
                JZ      ABOUTCR
                CMP     EAX,21H
                JZ      ABOUTDR
                CMP     EAX,24H
                JZ      ABOUTTR
                JMP     NOTRECOVER
ABOUTCR:
                MOV     ECX,0H
                MOV     EDX,1FH
                JMP     RUNINSTRUCTION
ABOUTDR:
                MOV     ECX,0H
                MOV     EDX,3FH
                JMP     RUNINSTRUCTION
ABOUTTR:
                MOV     ECX,30H
                MOV     EDX,3FH
RUNINSTRUCTION:
                MOV     EAX,ES:[EBX]
                SHR     EAX,16
                AND     EAX,03FH
                CMP     EAX,ECX
                JB      NOTRECOVER
                CMP     EAX,EDX
                JA      NOTRECOVER
                MOV     AX,SEL_ISR       ;The alias of ISR
                MOV     GS,AX
                MOV     EAX,ES:[EBX]
                MOV     BYTE PTR GS:THEINSTRUCTION,AL
                SHR     EAX,8
                MOV     BYTE PTR GS:THEINSTRUCTION+1,AL
                SHR     EAX,8
                MOV     BYTE PTR GS:THEINSTRUCTION+2,AL
                MOV     EAX,[ESP+30H+4]
                ADD     EAX,3
                MOV     [ESP+30H+4],EAX

                POP     GS
                POP     FS
                POP     ES
                POPAD
THEINSTRUCTION  LABEL
                DB      0,0,0
                ADD     ESP,8
                IRETD

SKIPINSTRUCTION:
                MOV     EAX,[ESP+30H+4]
                ADD     EAX,3
                MOV     [ESP+30H+4],EAX
                POP     GS
                POP     FS
                POP     ES
                POPAD
                ADD     ESP,8
                IRETD

;****************************************************************************
;*  here handle the I/O instruction,Analyse                                 *
;****************************************************************************

NOTMOV:
                CMP     BYTE PTR FS:EXEFLAG,0H
                JZ      EMULATEDOG    ;IS 1, analyse the soft dog
                CMP     EAX,0ECH      ;IN AL,DX
                JZ      INAL_DX
                CMP     EAX,0EDH      ;IN AX,DX
                JZ      INAX_DX
                CMP     EAX,0E4H      ;IN AL,XX
                JZ      INAL_XX
                CMP     EAX,0E5H      ;IN AX,XX
                JZ      INAX_XX
                CMP     EAX,0EEH      ;OUT DX,AL
                JZ      OUTDX_AL
                CMP     EAX,0EFH      ;OUT DX,AX
                JZ      OUTDX_AX
                CMP     EAX,0E6H      ;OUT XX,AL
                JZ      OUTXX_AL
                CMP     EAX,0E7H      ;OUT XX,AX
                JZ      OUTXX_AX
                CMP     EAX,06CH      ;INSB
                JZ      ABOUTINSB
                CMP     EAX,06DH      ;INSW
                JZ      ABOUTINSW
                CMP     EAX,6EH       ;OUTSB
                JZ      ABOUTOUTSB
                CMP     EAX,6FH       ;OUTSW
                JZ      ABOUTOUTSW
                CMP     EAX,0F3H      ;REP
                JZ      ABOUTREP
                JMP     NOTRECOVER
INAL_DX:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EDX,[ESP+20H] ;OLD EDX
                IN      AL,DX
                MOV     ESI,GS:PORTDATAADDR
                MOV     EBX,ES:[ESI+2]
                MOV     BYTE PTR ES:[ESI+EBX+6],0    ;in byte
                MOV     WORD PTR ES:[ESI+EBX+7],DX   ;port no
                MOV     BYTE PTR ES:[ESI+EBX+9],AL   ;value
                ADD     EBX,4
                MOV     ES:[ESI+2],EBX                   ;counter
                MOV     BYTE PTR [ESP+28H],AL
                MOV     EAX,[ESP+30H+4]                 ;to next instruction
                INC     EAX
                MOV     [ESP+30H+4],EAX
                JMP     IOINSTRUCTIONOVER
INAX_DX:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EDX,[ESP+20H] ;OLD EDX
                IN      AX,DX
                MOV     ESI,GS:PORTDATAADDR
                MOV     EBX,ES:[ESI+2]
                MOV     BYTE PTR ES:[ESI+EBX+6],1    ;in word
                MOV     WORD PTR ES:[ESI+EBX+7],DX   ;port no
                MOV     WORD PTR ES:[ESI+EBX+9],AX   ;value
                ADD     EBX,5
                MOV     ES:[ESI+2],EBX                   ;counter
                MOV     WORD PTR [ESP+28H],AX
                MOV     EAX,[ESP+30H+4]                 ;to next instruction
                INC     EAX
                MOV     [ESP+30H+4],EAX
                JMP     IOINSTRUCTIONOVER
INAL_XX:
                MOV     EDX,ES:[EBX]          ;CURRENT INSTRUCTION
                SHR     EDX,8
                AND     EDX,0FFH
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                IN      AL,DX
                MOV     ESI,GS:PORTDATAADDR
                MOV     EBX,ES:[ESI+2]
                MOV     BYTE PTR ES:[ESI+EBX+6],0     ;in byte
                MOV     WORD PTR ES:[ESI+EBX+7],DX    ;port no
                MOV     BYTE PTR ES:[ESI+EBX+9],AL    ;value
                ADD     EBX,4
                MOV     ES:[ESI+2],EBX                    ;counter
                MOV     BYTE PTR [ESP+28H],AL
                MOV     EAX,[ESP+30H+4]                 ;to next instruction
                ADD     EAX,2
                MOV     [ESP+30H+4],EAX
                JMP     IOINSTRUCTIONOVER
INAX_XX:
                MOV     EDX,ES:[EBX]          ;CURRENT INSTRUCTION
                SHR     EDX,8
                AND     EDX,0FFH
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                IN      AX,DX
                MOV     ESI,GS:PORTDATAADDR
                MOV     EBX,ES:[ESI+2]
                MOV     BYTE PTR ES:[ESI+EBX+6],1    ;in word
                MOV     WORD PTR ES:[ESI+EBX+7],DX   ;port no
                MOV     WORD PTR ES:[ESI+EBX+9],AX   ;value
                ADD     EBX,5
                MOV     ES:[ESI+2],EBX                   ;counter
                MOV     WORD PTR [ESP+28H],AX
                MOV     EAX,[ESP+30H+4]                  ;to next instruction
                ADD     EAX,2
                MOV     [ESP+30H+4],EAX
                JMP     IOINSTRUCTIONOVER
OUTDX_AL:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EDX,[ESP+20H] ;OLD EDX
                MOV     AL,BYTE PTR [ESP+28H]   ;OLD EAX
                OUT     DX,AL
                MOV     ESI,GS:PORTDATAADDR
                MOV     EBX,ES:[ESI+2]
                MOV     BYTE PTR ES:[ESI+EBX+6],10H   ;out byte
                MOV     WORD PTR ES:[ESI+EBX+7],DX    ;port no
                MOV     BYTE PTR ES:[ESI+EBX+9],AL    ;value
                ADD     EBX,4
                MOV     ES:[ESI+2],EBX                    ;counter
                MOV     EAX,[ESP+30H+4]                    ;to next instruction
                INC     EAX
                MOV     [ESP+30H+4],EAX
                JMP     IOINSTRUCTIONOVER
OUTDX_AX:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EDX,[ESP+20H]           ;OLD EDX
                MOV     AX,WORD PTR [ESP+28H]   ;OLD EAX
                OUT     DX,AX
                MOV     ESI,GS:PORTDATAADDR
                MOV     EBX,ES:[ESI+2]
                MOV     BYTE PTR ES:[ESI+EBX+6],11H  ;out word
                MOV     WORD PTR ES:[ESI+EBX+7],DX   ;portno
                MOV     WORD PTR ES:[ESI+EBX+9],AX   ;value
                ADD     EBX,5
                MOV     ES:[ESI+2],EBX                   ;counter
                MOV     EAX,[ESP+30H+4]                   ;to next instruction
                INC     EAX
                MOV     [ESP+30H+4],EAX
                JMP     IOINSTRUCTIONOVER
OUTXX_AL:
                MOV     EDX,ES:[EBX]          ;CURRENT INSTRUCTION
                SHR     EDX,8
                AND     EDX,0FFH
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     AL,BYTE PTR [ESP+28H] ;OLD EAX
                OUT     DX,AL
                MOV     ESI,GS:PORTDATAADDR
                MOV     EBX,ES:[ESI+2]
                MOV     BYTE PTR ES:[ESI+EBX+6],10H ;word byte
                MOV     WORD PTR ES:[ESI+EBX+7],DX  ;port no
                MOV     BYTE PTR ES:[ESI+EBX+9],AL  ;value
                ADD     EBX,4
                MOV     ES:[ESI+2],EBX
                MOV     EAX,[ESP+30H+4]                 ;to next instruction
                ADD     EAX,2
                MOV     [ESP+30H+4],EAX
                JMP     IOINSTRUCTIONOVER
OUTXX_AX:
                MOV     EDX,ES:[EBX]          ;CURRENT INSTRUCTION
                SHR     EDX,8
                AND     EDX,0FFH
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     AX,WORD PTR [ESP+28H]  ;OLD EAX
                OUT     DX,AX
                MOV     ESI,GS:PORTDATAADDR
                MOV     EBX,ES:[ESI+2]
                MOV     BYTE PTR ES:[ESI+EBX+6],11H  ;out word
                MOV     WORD PTR ES:[ESI+EBX+7],DX   ;port no
                MOV     WORD PTR ES:[ESI+EBX+9],AX   ;value
                ADD     EBX,5
                MOV     ES:[ESI+2],EBX
                MOV     EAX,[ESP+30H+4]                 ;to next instruction
                ADD     EAX,2
                MOV     [ESP+30H+4],EAX
                JMP     IOINSTRUCTIONOVER
ABOUTINSB:
                MOV     EAX,[ESP+44H+4]      ;OLD ES
                SHL     EAX,4
                MOV     EDI,[ESP+0CH]        ;OLD EDI
                ADD     EDI,EAX
                MOV     EDX,[ESP+20H]        ;OLD DX
                CLD
                MOV     EAX,[ESP+38H+4]      ;OLD FLAG,DF
                AND     EAX,400H
                JZ      NOTCHANGE1
                STD
                INSB
                MOV     CL,BYTE PTR ES:[EDI+1]
                MOV     EAX,[ESP+0CH]
                DEC     EAX
                AND     EAX,0FFFFH
                MOV     [ESP+0CH],EAX
                JMP     INSBOVER
NOTCHANGE1:
                INSB
                MOV     CL,BYTE PTR ES:[EDI-1]
                MOV     EAX,[ESP+0CH]
                INC     EAX
                AND     EAX,0FFFFH
                MOV     [ESP+0CH],EAX
INSBOVER:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     ESI,GS:PORTDATAADDR
                MOV     EBX,ES:[ESI+2]
                MOV     BYTE PTR ES:[ESI+EBX+6],0H  ;in byte
                MOV     WORD PTR ES:[ESI+EBX+7],DX  ;port no
                MOV     BYTE PTR ES:[ESI+EBX+9],CL  ;value
                ADD     EBX,4
                MOV     ES:[ESI+2],EBX
                MOV     EAX,[ESP+30H+4]                ;to next instruction
                INC     EAX
                MOV     [ESP+30H+4],EAX
                JMP     IOINSTRUCTIONOVER
ABOUTINSW:
                MOV     EAX,[ESP+44H+4]      ;OLD ES
                SHL     EAX,4
                MOV     EDI,[ESP+0CH]        ;OLD EDI
                ADD     EDI,EAX
                MOV     EDX,[ESP+20H]        ;OLD DX
                CLD
                MOV     EAX,[ESP+38H+4]      ;OLD FLAG,DF
                AND     EAX,400H
                JZ      NOTCHANGE2
                STD
                INSW
                MOV     CX,WORD PTR ES:[EDI+2]
                MOV     EAX,[ESP+0CH]
                SUB     EAX,2
                MOV     [ESP+0CH],EAX
                JMP     INSWOVER
NOTCHANGE2:
                INSW
                MOV     CX,WORD PTR ES:[EDI-2]
                MOV     EAX,[ESP+0CH]
                ADD     EAX,2
                MOV     [ESP+0CH],EAX
INSWOVER:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     ESI,GS:PORTDATAADDR
                MOV     EBX,ES:[ESI+2]
                MOV     BYTE PTR ES:[ESI+EBX+4],1H  ;in word
                MOV     WORD PTR ES:[ESI+EBX+5],DX  ;port no
                MOV     WORD PTR ES:[ESI+EBX+7],CX  ;value
                ADD     EBX,5
                MOV     ES:[ESI+2],EBX
                MOV     EAX,[ESP+30H+4]                ;to next instruction
                INC     EAX
                MOV     [ESP+30H+4],EAX
                JMP     IOINSTRUCTIONOVER

ABOUTOUTSB:
                MOV     EAX,[ESP+48H+4]      ;OLD DS
                SHL     EAX,4
                MOV     ESI,[ESP+10H]        ;OLD ESI
                ADD     ESI,EAX
                MOV     EDX,[ESP+20H]        ;OLD DX
                CLD
                MOV     EAX,[ESP+38H+4]      ;OLD FLAG,DF
                AND     EAX,400H
                JZ      NOTCHANGE3
                STD
                OUTSB
                MOV     CL,BYTE PTR ES:[ESI+1]
                MOV     EAX,[ESP+0CH]
                DEC     EAX
                AND     EAX,0FFFFH
                MOV     [ESP+0CH],EAX
                JMP     OUTSBOVER
NOTCHANGE3:
                OUTSB
                MOV     CL,BYTE PTR ES:[ESI-1]
                MOV     EAX,[ESP+0CH]
                INC     EAX
                AND     EAX,0FFFFH
                MOV     [ESP+0CH],EAX
OUTSBOVER:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     ESI,GS:PORTDATAADDR
                MOV     EBX,ES:[ESI+2]
                MOV     BYTE PTR ES:[ESI+EBX+6],10H  ;out byte
                MOV     WORD PTR ES:[ESI+EBX+7],DX   ;port no
                MOV     BYTE PTR ES:[ESI+EBX+9],CL   ;value
                ADD     EBX,4
                MOV     ES:[ESI+2],EBX
                MOV     EAX,[ESP+30H+4]
                INC     EAX
                MOV     [ESP+30H+4],EAX                 ;next instruction
                JMP     IOINSTRUCTIONOVER

ABOUTOUTSW:
                MOV     EAX,[ESP+48H+4]      ;OLD DS
                SHL     EAX,4
                MOV     ESI,[ESP+10H]        ;OLD ESI
                ADD     ESI,EAX
                MOV     EDX,[ESP+20H]        ;OLD DX
                CLD
                MOV     EAX,[ESP+38H+4]      ;OLD FLAG,DF
                AND     EAX,400H
                JZ      NOTCHANGE4
                STD
                OUTSW
                MOV     CX,WORD PTR ES:[ESI+2]
                MOV     EAX,[ESP+0CH]
                SUB     EAX,2
                AND     EAX,0FFFFH
                MOV     [ESP+0CH],EAX
                JMP     OUTSWOVER
NOTCHANGE4:
                OUTSW
                MOV     CX,WORD PTR ES:[ESI-2]
                MOV     EAX,[ESP+0CH]
                ADD     EAX,2
                AND     EAX,0FFFFH
                MOV     [ESP+0CH],EAX
OUTSWOVER:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     ESI,GS:PORTDATAADDR
                MOV     EBX,ES:[ESI+2]
                MOV     BYTE PTR ES:[ESI+EBX+6],11H  ;out word
                MOV     WORD PTR ES:[ESI+EBX+7],DX   ;port no
                MOV     WORD PTR ES:[ESI+EBX+9],CX   ;value
                ADD     EBX,5
                MOV     ES:[ESI+2],EBX
                MOV     EAX,[ESP+30H+4]
                INC     EAX
                MOV     [ESP+30H+4],EAX                 ;next instruction
                JMP     IOINSTRUCTIONOVER

ABOUTREP:
                MOV     EAX,GS:[EBX]
                SHR     EAX,8
                AND     EAX,0FFH
                CMP     EAX,6CH
                JZ      ABOUTREPINSB
                CMP     EAX,6DH
                JZ      ABOUTREPINSW
                CMP     EAX,6EH
                JZ      ABOUTREPOUTSB
                CMP     EAX,6FH
                JZ      ABOUTREPOUTSW
                JMP     NOTRECOVER
ABOUTREPINSB:
ABOUTREPINSW:
ABOUTREPOUTSB:
ABOUTREPOUTSW:
                JMP     IOINSTRUCTIONOVER

IOINSTRUCTIONOVER:
                POP     GS
                POP     FS
                POP     ES
                POPAD
                ADD     ESP,8
                IRETD

;****************************************************************************
;*  here handle the I/O instruction,Emulation                               *
;****************************************************************************

EMULATEDOG:
                CMP     EAX,0ECH      ;IN AL,DX
                JZ      EMINAL_DX
                CMP     EAX,0EDH      ;IN AX,DX
                JZ      EMINAX_DX
                CMP     EAX,0E4H      ;IN AL,XX
                JZ      EMINAL_XX
                CMP     EAX,0E5H      ;IN AX,XX
                JZ      EMINAX_XX
                CMP     EAX,0EEH      ;OUT DX,AL
                JZ      EMOUTDX_AL
                CMP     EAX,0EFH      ;OUT DX,AX
                JZ      EMOUTDX_AX
                CMP     EAX,0E6H      ;OUT XX,AL
                JZ      EMOUTXX_AL
                CMP     EAX,0E7H      ;OUT XX,AX
                JZ      EMOUTXX_AX
                CMP     EAX,06CH      ;INSB
                JZ      EMABOUTINSB
                CMP     EAX,06DH      ;INSW
                JZ      EMABOUTINSW
                CMP     EAX,6EH       ;OUTSB
                JZ      EMABOUTOUTSB
                CMP     EAX,6FH       ;OUTSW
                JZ      EMABOUTOUTSW
                CMP     EAX,0F3H      ;REP
                JZ      EMABOUTREP
                JMP     NOTRECOVER
EMINAL_DX:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EDX,[ESP+20H]           ;OLD EDX
                MOV     AL,0                    ;IN  BYTE
                CALL32F SEL_ICODE,CHECKINPORT
                JC      NOTEMULATION
                MOV     BYTE PTR [ESP+28H],AL   ;EMULATION VALUE
                MOV     EAX,[ESP+30H+4]         ;to next instruction
                INC     EAX
                MOV     [ESP+30H+4],EAX
                JMP     EMIOINSTRUCTIONOVER
EMINAX_DX:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EDX,[ESP+20H]           ;OLD EDX
                MOV     AL,1                    ;IN WORD
                CALL32F SEL_ICODE,CHECKINPORT
                JC      NOTEMULATION
                MOV     WORD PTR [ESP+28H],AX   ;EMULATION VALUE
                MOV     EAX,[ESP+30H+4]         ;to next instruction
                INC     EAX
                MOV     [ESP+30H+4],EAX
                JMP     EMIOINSTRUCTIONOVER
EMINAL_XX:
                MOV     EDX,ES:[EBX]            ;CURRENT INSTRUCTION
                SHR     EDX,8
                AND     EDX,0FFH                ;PORT NO
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     AL,0                    ;IN BYTE
                CALL32F SEL_ICODE,CHECKINPORT
                JC      NOTEMULATION
                MOV     BYTE PTR [ESP+28H],AL
                MOV     EAX,[ESP+30H+4]         ;to next instruction
                ADD     EAX,2
                MOV     [ESP+30H+4],EAX
                JMP     EMIOINSTRUCTIONOVER
EMINAX_XX:
                MOV     EDX,ES:[EBX]            ;CURRENT INSTRUCTION
                SHR     EDX,8
                AND     EDX,0FFH                ;PORT NO
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     AL,1                    ;IN WORD
                CALL32F SEL_ICODE,CHECKINPORT
                JC      NOTEMULATION
                MOV     WORD PTR [ESP+28H],AX   ;EMULATION VALUE
                MOV     EAX,[ESP+30H+4]         ;to next instruction
                ADD     EAX,2
                MOV     [ESP+30H+4],EAX
                JMP     EMIOINSTRUCTIONOVER
EMOUTDX_AL:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EDX,[ESP+20H]           ;OLD EDX
                MOV     AL,BYTE PTR [ESP+28H]   ;OLD EAX
                MOV     BL,10H                  ;OUT BYTE
                CALL32F SEL_ICODE,CHECKOUTPORT
                JC      NOTEMULATION
                MOV     EAX,[ESP+30H+4]                 ;to next instruction
                INC     EAX
                MOV     [ESP+30H+4],EAX
                JMP     EMIOINSTRUCTIONOVER
EMOUTDX_AX:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EDX,[ESP+20H]           ;OLD EDX
                MOV     AX,WORD PTR [ESP+28H]   ;OLD EAX
                MOV     BL,11H                  ;OUT WORD
                CALL32F SEL_ICODE,CHECKOUTPORT
                JC      NOTEMULATION
                MOV     EAX,[ESP+30H+4]                 ;to next instruction
                INC     EAX
                MOV     [ESP+30H+4],EAX
                JMP     EMIOINSTRUCTIONOVER
EMOUTXX_AL:
                MOV     EDX,ES:[EBX]            ;CURRENT INSTRUCTION
                SHR     EDX,8
                AND     EDX,0FFH                ;OLD EDX
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     AL,BYTE PTR [ESP+28H]   ;OLD EAX
                MOV     BL,10H                  ;OUT BYTE
                CALL32F SEL_ICODE,CHECKOUTPORT
                JC      NOTEMULATION
                MOV     EAX,[ESP+30H+4]         ;to next instruction
                ADD     EAX,2
                MOV     [ESP+30H+4],EAX
                JMP     EMIOINSTRUCTIONOVER
EMOUTXX_AX:
                MOV     EDX,ES:[EBX]            ;CURRENT INSTRUCTION
                SHR     EDX,8
                AND     EDX,0FFH
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     AX,WORD PTR [ESP+28H]   ;OLD EAX
                MOV     BL,11H                  ;OUT WORD
                CALL32F SEL_ICODE,CHECKOUTPORT
                JC      NOTEMULATION
                MOV     EAX,[ESP+30H+4]         ;to next instruction
                ADD     EAX,2
                MOV     [ESP+30H+4],EAX
                JMP     EMIOINSTRUCTIONOVER
EMABOUTINSB:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EAX,[ESP+44H+4]         ;OLD ES
                SHL     EAX,4
                MOV     EDI,[ESP+0CH]           ;OLD EDI
                ADD     EDI,EAX
                MOV     EDX,[ESP+20H]           ;OLD DX
                MOV     EAX,[ESP+38H+4]         ;OLD FLAG,DF
                AND     EAX,400H
                JZ      EMNOTCHANGE1
                MOV     AL,0                    ;INBYTE
                CALL32F SEL_ICODE,CHECKINPORT
                JC      NOTEMULATION
                MOV     BYTE PTR ES:[EDI],AL    ;EMULATE DATA
                MOV     EAX,[ESP+0CH]
                DEC     EAX
                AND     EAX,0FFFFH
                MOV     [ESP+0CH],EAX           ;change EDI
                JMP     EMINSBOVER
EMNOTCHANGE1:
                MOV     AL,0                    ;IN BYTE
                CALL32F SEL_ICODE,CHECKINPORT
                JC      NOTEMULATION
                MOV     BYTE PTR ES:[EDI],AL
                MOV     EAX,[ESP+0CH]
                INC     EAX
                AND     EAX,0FFFFH
                MOV     [ESP+0CH],EAX           ;change EDI
EMINSBOVER:
                MOV     EAX,[ESP+30H+4]                ;to next instruction
                INC     EAX
                MOV     [ESP+30H+4],EAX
                JMP     EMIOINSTRUCTIONOVER
EMABOUTINSW:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EAX,[ESP+44H+4]         ;OLD ES
                SHL     EAX,4
                MOV     EDI,[ESP+0CH]           ;OLD EDI
                ADD     EDI,EAX
                MOV     EDX,[ESP+20H]           ;OLD DX
                MOV     EAX,[ESP+38H+4]         ;OLD FLAG,DF
                AND     EAX,400H
                JZ      EMNOTCHANGE2
                MOV     AL,1                    ;IN WORD
                CALL32F SEL_ICODE,CHECKINPORT
                MOV     WORD PTR ES:[EDI],AX
                MOV     EAX,[ESP+0CH]
                SUB     EAX,2
                MOV     [ESP+0CH],EAX           ;CHANGE EDI
                JMP     EMINSWOVER
EMNOTCHANGE2:
                MOV     AL,1                    ;IN WORD
                CALL32F SEL_ICODE,CHECKINPORT
                MOV     WORD PTR ES:[EDI],AX
                MOV     EAX,[ESP+0CH]
                ADD     EAX,2
                MOV     [ESP+0CH],EAX            ;CHANGE EDI
EMINSWOVER:
                MOV     EAX,[ESP+30H+4]          ;to next instruction
                INC     EAX
                MOV     [ESP+30H+4],EAX
                JMP     EMIOINSTRUCTIONOVER

EMABOUTOUTSB:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EAX,[ESP+48H+4]      ;OLD DS
                SHL     EAX,4
                MOV     ESI,[ESP+10H]        ;OLD ESI
                ADD     ESI,EAX
                MOV     EDX,[ESP+20H]        ;OLD DX
                MOV     EAX,[ESP+38H+4]      ;OLD FLAG,DF
                AND     EAX,400H
                JZ      EMNOTCHANGE3
                MOV     AL,BYTE PTR ES:[ESI]
                MOV     BL,10                ;OUT BYTE
                CALL32F SEL_ICODE,CHECKOUTPORT
                JC      NOTEMULATION
                MOV     EAX,[ESP+10H]
                DEC     EAX
                AND     EAX,0FFFFH
                MOV     [ESP+10H],EAX        ;CHANGE ESI
                JMP     EMOUTSBOVER
EMNOTCHANGE3:
                MOV     AL,BYTE PTR ES:[ESI]
                MOV     BL,10                ;OUT BYTE
                CALL32F SEL_ICODE,CHECKOUTPORT
                JC      NOTEMULATION
                MOV     EAX,[ESP+10H]
                INC     EAX
                AND     EAX,0FFFFH
                MOV     [ESP+10H],EAX        ;CHANGE EDI
EMOUTSBOVER:
                MOV     EAX,[ESP+30H+4]       ;OLD EIP
                INC     EAX
                MOV     [ESP+30H+4],EAX       ;next instruction
                JMP     EMIOINSTRUCTIONOVER

EMABOUTOUTSW:
                MOV     AX,SEL_UDATA
                MOV     GS,AX
                MOV     EAX,[ESP+48H+4]      ;OLD DS
                SHL     EAX,4
                MOV     ESI,[ESP+10H]        ;OLD ESI
                ADD     ESI,EAX
                MOV     EDX,[ESP+20H]        ;OLD DX
                MOV     EAX,[ESP+38H+4]      ;OLD FLAG,DF
                AND     EAX,400H
                JZ      EMNOTCHANGE4
                MOV     AX,WORD PTR ES:[ESI]
                MOV     BL,11H                ;OUT WORD
                CALL32F SEL_ICODE,CHECKOUTPORT
                JC      NOTEMULATION
                MOV     EAX,[ESP+10H]
                SUB     EAX,2
                AND     EAX,0FFFFH
                MOV     [ESP+10H],EAX         ;CHANGE ESI
                JMP     EMOUTSWOVER
EMNOTCHANGE4:
                MOV     AX,WORD PTR ES:[ESI]
                MOV     BL,11H
                CALL32F SEL_ICODE,CHECKOUTPORT
                JC      NOTEMULATION
                MOV     EAX,[ESP+10H]
                ADD     EAX,2
                AND     EAX,0FFFFH
                MOV     [ESP+10H],EAX         ;CHANGE ESI
EMOUTSWOVER:
                MOV     EAX,[ESP+30H+4]       ;OLD EIP
                INC     EAX
                MOV     [ESP+30H+4],EAX       ;next instruction
                JMP     EMIOINSTRUCTIONOVER

EMABOUTREP:
                MOV     EAX,GS:[EBX]
                SHR     EAX,8
                AND     EAX,0FFH
                CMP     EAX,6CH
                JZ      EMABOUTREPINSB
                CMP     EAX,6DH
                JZ      EMABOUTREPINSW
                CMP     EAX,6EH
                JZ      EMABOUTREPOUTSB
                CMP     EAX,6FH
                JZ      EMABOUTREPOUTSW
                JMP     NOTRECOVER
EMABOUTREPINSB:
EMABOUTREPINSW:
EMABOUTREPOUTSB:
EMABOUTREPOUTSW:
                JMP     EMIOINSTRUCTIONOVER

EMIOINSTRUCTIONOVER:
                POP     GS
                POP     FS
                POP     ES
                POPAD
                ADD     ESP,8
                IRETD
NOTGP:
                MOV     EAX,[ESP+30H]
                ADD     EAX,2
                MOV     [ESP+30H],EAX
                POP     GS
                POP     FS
                POP     ES
                POPAD
                ADD     ESP,4
                IRETD


NOTEMULATION:
                MOV     EAX,GS:CURBLOCKADDR
                CALL32F SEL_CODE32,HEXOUT4
                MOV     EAX,GS:BYTESCOUNTER
                CALL32F SEL_CODE32,HEXOUT4



NOTRECOVER:
                MOV     EAX,[ESP+30H+4]
                MOV     EBX,[ESP+34H+4]
                SHL     EBX,4
                ADD     EBX,EAX
                MOV     EAX,ES:[EBX]
                CALL32F SEL_CODE32,HEXOUT4
                POP     GS
                POP     FS
                POP     ES
                POPAD

                PUSHAD
                PUSH    GS
                PUSH    FS
                PUSH    DS
                PUSH    ES
                MOV     AX,SEL_VIDEO
                MOV     ES,AX
                MOV     AX,CS
                MOV     DS,AX
;do dump
                MOV     ECX,4
INTL1:
                MOV     AL,[RTABLE-1+ECX]
                CALL32F SEL_CODE32,OUCH
                MOV     AL,'S'
                CALL32F SEL_CODE32,OUCH
                MOV     AL,'='
                CALL32F SEL_CODE32,OUCH
                POP     EAX
                CALL32F SEL_CODE32,HEXOUT2
                PUSH    ECX
                MOV     ECX,6
LSP1:           MOV     AL,' '
                CALL32F SEL_CODE32,OUCH
                LOOP    LSP1
                POP     ECX
                LOOP    INTL1
                CALL32F SEL_CODE32,CRLF
                XOR     ECX,ECX
INTL2:          CMP     CL,5
                JNZ     SHORT NOCRINT
                CALL32F SEL_CODE32,CRLF
NOCRINT:
                MOV     AL,'E'
                CALL32F SEL_CODE32,OUCH
                MOV     AL,[GTABLE+ECX*2]
                CALL32F SEL_CODE32,OUCH
                MOV     AL,[GTABLE+1+ECX*2]
                CALL32F SEL_CODE32,OUCH
                MOV     AL,'='
                CALL32F SEL_CODE32,OUCH
                POP     EAX
                CALL32F SEL_CODE32,HEXOUT4
                MOV     AL,' '
                CALL32F SEL_CODE32,OUCH
                INC     CL
                CMP     CL,8
                JNE     SHORT INTL2
                MOV     EBX,OFFSET MESSAGEA
                CALL32F SEL_CODE32,MESSOUT
                POP     EAX
                CALL32F SEL_CODE32,HEXOUT
                MOV     EBX,OFFSET TASKM
                CALL32F SEL_CODE32,MESSOUT
                STR     AX
                CALL32F SEL_CODE32,HEXOUT2
                CALL32F SEL_CODE32,CRLF
;stack dump
                MOV     AX,SS
                MOVZX   EAX,AX
                LSL     EDX,EAX
                JNZ     SHORT INTABT
                MOV     EBX,OFFSET STKM
                CALL32F SEL_CODE32,MESSOUT
                XOR     CL,CL
                MOV     SI,64
INTL3:          CMP     ESP,EDX
                JAE     SHORT INTABT
                DEC     SI
                JZ      SHORT INTABT
                TEST    CL,7
                JNZ     SHORT NOSCR
                CALL32F SEL_CODE32,CRLF
NOSCR:
                POP     EAX
                CALL32F SEL_CODE32,HEXOUT4
                INC     CL
                MOV     AL,' '
                CALL32F SEL_CODE32,OUCH
                JMP     SHORT INTL3
INTABT:
;check for memory dump request
                MOV     AX,SEL_DATA
                MOV     DS,AX
                ASSUME  DS:DAT32
                MOV     AX,WORD PTR DUMP_SEG
                OR      AX,AX
                JZ      SHORT NOMEMDUMP
;come here to do memory dump
                CALL32F SEL_CODE32,CRLF
                PUSH    DS
                PUSH    CS
                POP     DS
                MOV     EBX,OFFSET MEMMESS
                CALL32F SEL_CODE32,MESSOUT
                CALL32F SEL_CODE32,CRLF
                POP     DS
                MOV     AX,WORD PTR DUMP_SEG
                MOV     ES,AX
                CALL32F SEL_CODE32,HEXOUT2
                MOV     AL,':'
                CALL32F SEL_CODE32,OUCH
                MOV     EDX,DUMP_OFF
                MOV     EAX,EDX
                CALL32F SEL_CODE32,HEXOUT4
                MOV     ECX,DUMP_CNT
DUMPLOOP:
                MOV     AL,' '
                CALL32F SEL_CODE32,OUCH
                MOV     EAX,ES:[EDX]    ;get word
                CALL32F SEL_CODE32,HEXOUT4
                ADD     EDX,4
                SUB     ECX,4
                JA      SHORT DUMPLOOP
                CALL32F SEL_CODE32,CRLF
NOMEMDUMP:
                MOV     AL,20H          ;Send EOI signal
                CMP     _PC386,0
                JNZ     SHORT NOTANAT2
                OUT     0A0H,AL
NOTANAT2:
                OUT     20H,AL          ;just in case hardware did it
                MOV     AL,7FH          ;return 7f to dos
                BACK2DOS
;*****************************************************************************
;* Here emulation input a data from a port                                   *
;*****************************************************************************

CHECKINPORT     PROC    FAR
                PUSH    EBX
                PUSH    ECX
                PUSH    EDX
                PUSH    ESI
                PUSH    EDI

                MOV     BYTE PTR GS:_CURTYPE,AL
                MOV     WORD PTR GS:_CURPORT,DX
                MOV     ESI,GS:CURBLOCKADDR
                MOV     EDI,ESI
                MOV     EBX,GS:BYTESCOUNTER
CMPINPORT:
                MOV     AL,BYTE PTR GS:_CURTYPE
                CMP     AL,BYTE PTR ES:[ESI+EBX+4]        ;DATA TYPE
                JNZ     INSEARCHNEXTBLOCK
                MOV     DX,WORD PTR GS:_CURPORT
                CMP     DX,WORD PTR ES:[ESI+EBX+5]        ;PORT NO
                JNZ     INSEARCHNEXTBLOCK
                MOV     GS:CURBLOCKADDR,ESI
                MOV     GS:BYTESCOUNTER,EBX
                CMP     AL,0
                JZ      INBYTE
                CMP     AL,1
                JZ      INWORD
                CMP     AL,2
                JZ      INDWORD
                MOV     AL,GS:_CURTYPE
                POP     EDI
                POP     ESI
                POP     EDX
                POP     ECX
                POP     EBX
                STC
                RETF
INBYTE:
                MOV     EAX,GS:BYTESCOUNTER
                ADD     EAX,4
                CMP     EAX,ES:[ESI]
                JB      INNOTCIRCLE1
                MOV     EAX,0
INNOTCIRCLE1:
                MOV     GS:BYTESCOUNTER,EAX
                MOV     AL,BYTE PTR ES:[ESI+EBX+7]
                POP     EDI
                POP     ESI
                POP     EDX
                POP     ECX
                POP     EBX
                CLC
                RETF
INWORD:
                MOV     EAX,GS:BYTESCOUNTER
                ADD     EAX,5
                CMP     EAX,ES:[ESI]
                JB      INNOTCIRCLE2
                MOV     EAX,0
INNOTCIRCLE2:
                MOV     GS:BYTESCOUNTER,EAX
                MOV     AX,WORD PTR ES:[ESI+EBX+7]
                POP     EDI
                POP     ESI
                POP     EDX
                POP     ECX
                POP     EBX
                CLC
                RETF
INDWORD:
                MOV     EAX,GS:BYTESCOUNTER
                ADD     EAX,7
                CMP     EAX,ES:[ESI]
                JB      INNOTCIRCLE3
                MOV     EAX,0
INNOTCIRCLE3:
                MOV     GS:BYTESCOUNTER,EAX
                MOV     EAX,DWORD PTR ES:[ESI+EBX+7]
                POP     EDI
                POP     ESI
                POP     EDX
                POP     ECX
                POP     EBX
                CLC
                RETF
INSEARCHNEXTBLOCK:
                MOV     EAX,ES:[EDI]
                ADD     EDI,EAX
                ADD     EDI,4
                MOV     EAX,ES:[EDI]
                CMP     EAX,EBX
                JB      INEMULATIONERROR
                MOV     EDX,EBX
                MOV     EBX,0
CMPINLOOP:
                CMP     EBX,EDX
                JZ      INEQUAL
                MOV     AL,ES:[ESI+EBX+4]
                CMP     AL,ES:[EDI+EBX+4]
                JNZ     INNOTEQUAL
                INC     EBX
                JMP     CMPINLOOP
INEQUAL:
                MOV     EBX,EDX
                MOV     ESI,EDI
                JMP     CMPINPORT
INNOTEQUAL:
                MOV     EBX,EDX
                JMP     INSEARCHNEXTBLOCK
INEMULATIONERROR:
                MOV     AL,GS:_CURTYPE
                POP     EDI
                POP     ESI
                POP     EDX
                POP     ECX
                POP     EBX
                STC
                RETF
CHECKINPORT     ENDP

;*****************************************************************************
;* Here emulation output a data to a port                                    *
;*****************************************************************************

CHECKOUTPORT    PROC    FAR
                PUSH    EBX
                PUSH    ECX
                PUSH    EDX
                PUSH    ESI
                PUSH    EDI

                MOV     BYTE PTR GS:_CURTYPE,BL          ;current type
                MOV     WORD PTR GS:_CURPORT,DX          ;current port
                MOV     DWORD PTR GS:_CURVALUE,EAX       ;current output value
                MOV     ESI,GS:CURBLOCKADDR
                MOV     EDI,ESI
                MOV     EBX,GS:BYTESCOUNTER
CMPOUTPORT:
                MOV     AL,BYTE PTR GS:_CURTYPE
                CMP     AL,BYTE PTR ES:[ESI+EBX+4]        ;DATA TYPE
                JNZ     OUTSEARCHNEXTBLOCK
                MOV     DX,WORD PTR GS:_CURPORT
                CMP     DX,WORD PTR ES:[ESI+EBX+5]        ;PORT NO
                JNZ     OUTSEARCHNEXTBLOCK
                CMP     AL,10H
                JZ      OUTBYTE
                CMP     AL,11H
                JZ      OUTWORD
                CMP     AL,12H
                JZ      OUTDWORD
                POP     EDI
                POP     ESI
                POP     EDX
                POP     ECX
                POP     EBX
                STC
                RETF
OUTBYTE:
                MOV     EAX,DWORD PTR GS:_CURVALUE
                CMP     AL,BYTE PTR ES:[ESI+EBX+7]
                JNZ     OUTSEARCHNEXTBLOCK
                MOV     GS:CURBLOCKADDR,ESI
                MOV     GS:BYTESCOUNTER,EBX
                MOV     EAX,GS:BYTESCOUNTER
                ADD     EAX,4
                CMP     EAX,ES:[ESI]
                JB      OUTNOTCIRCLE1
                MOV     EAX,0
OUTNOTCIRCLE1:
                MOV     GS:BYTESCOUNTER,EAX
                POP     EDI
                POP     ESI
                POP     EDX
                POP     ECX
                POP     EBX
                CLC
                RETF
OUTWORD:
                MOV     EAX,DWORD PTR GS:_CURVALUE
                CMP     AX,WORD PTR ES:[ESI+EBX+7]
                JNZ     OUTSEARCHNEXTBLOCK
                MOV     GS:CURBLOCKADDR,ESI
                MOV     GS:BYTESCOUNTER,EBX
                MOV     EAX,GS:BYTESCOUNTER
                ADD     EAX,5
                CMP     EAX,ES:[ESI]
                JB      OUTNOTCIRCLE2
                MOV     EAX,0
OUTNOTCIRCLE2:
                MOV     GS:BYTESCOUNTER,EAX
                POP     EDI
                POP     ESI
                POP     EDX
                POP     ECX
                POP     EBX
                CLC
                RETF
OUTDWORD:
                MOV     EAX,DWORD PTR GS:_CURVALUE
                CMP     EAX,DWORD PTR ES:[ESI+EBX+7]
                JNZ     OUTSEARCHNEXTBLOCK
                MOV     GS:CURBLOCKADDR,ESI
                MOV     GS:BYTESCOUNTER,EBX
                MOV     EAX,GS:BYTESCOUNTER
                ADD     EAX,7
                CMP     EAX,ES:[ESI]
                JB      OUTNOTCIRCLE3
                MOV     EAX,0
OUTNOTCIRCLE3:
                MOV     GS:BYTESCOUNTER,EAX
                POP     EDI
                POP     ESI
                POP     EDX
                POP     ECX
                POP     EBX
                CLC
                RETF
OUTSEARCHNEXTBLOCK:
                MOV     EAX,ES:[EDI]
                ADD     EDI,EAX
                ADD     EDI,4
                MOV     EAX,ES:[EDI]
                CMP     EAX,EBX
                JB      OUTEMULATIONERROR
                MOV     EDX,EBX
                MOV     EBX,0
CMPOUTLOOP:
                CMP     EBX,EDX
                JZ      OUTEQUAL
                MOV     AL,ES:[ESI+EBX+4]
                CMP     AL,ES:[EDI+EBX+4]
                JNZ     OUTNOTEQUAL
                INC     EBX
                JMP     CMPOUTLOOP
OUTEQUAL:
                MOV     EBX,EDX
                MOV     ESI,EDI
                JMP     CMPOUTPORT
OUTNOTEQUAL:
                MOV     EBX,EDX
                JMP     OUTSEARCHNEXTBLOCK
OUTEMULATIONERROR:
                POP     EDI
                POP     ESI
                POP     EDX
                POP     ECX
                POP     EBX
                STC
                RETF
CHECKOUTPORT    ENDP


ISREND          EQU     $
ISR             ENDS

ISRLEN          EQU     (ISREND-ISRBEG)-1
IDTLEN          EQU     (IDTEND-IDTBEG)-1
INTLEN          EQU     (INT1-INT0)
                PUBLIC  ISRLEN,IDTLEN,INT0,INT1
                PUBLIC  IDTBEG,INTLEN,MESSAGEA
                END

